/*
    BFM, the Brutal File Manager. A 3D fps filemanager
    Copyright (C) 2003  Henrik Johansson and Daniel Forchheimer

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
    This file is almost entirely based on Andrew Davisons file
    QuadParticles.java from his particle system example found at
    http://fivedots.coe.psu.ac.th/~ad/jg/. If you're into Java3D and/or Java
    programming this is a great page, you should really check it out.

    A big thank to Andrew for letting us use his code!
*/

import javax.media.j3d.*;
import javax.vecmath.*;
import com.sun.j3d.utils.image.*;
import java.net.URL;

public class Blood extends OrientedShape3D {
	private static final float QUAD_LEN = 0.5f;
	private static final float GRAVITY = 9.8f;
	private static final float TIMESTEP = 0.05f;
	private static final float XZ_VELOCITY = 2.0f; 
	private static final float Y_VELOCITY = 8.0f; 
	private static final float DELTA = 0.05f; // for randomly adjusting the coords

	private QuadArray quadParts; // Geometry holding the coords, etc.

	private float[] cs, vels, accs, norms; // we must use floats arrays
	private float[] tcoords;

	// starting points for a quad, counter-clockewise from bottom-left
	private float[] p1 = {-QUAD_LEN/2, 0.0f, 0.0f};
	private float[] p2 = {QUAD_LEN/2, 0.0f, 0.0f};
	private float[] p3 = {QUAD_LEN/2, QUAD_LEN, 0.0f};
	private float[] p4 = {-QUAD_LEN/2, QUAD_LEN, 0.0f};
	
	private int numPoints;

	public Blood(int nps) {
		if (nps%4 != 0)// not a multiple of 4
			nps = ((int)((nps+4)/4))*4; // round up to nearest multiple
		numPoints = nps;

		// rotate about the y-axis to follow the viewer
		setAlignmentAxis( 0.0f, 1.0f, 0.0f);

		// BY_REFERENCE QuadArray
		quadParts = new QuadArray(numPoints, 
		                GeometryArray.COORDINATES | 
		                GeometryArray.TEXTURE_COORDINATE_2 |
		                GeometryArray.NORMALS |
		                GeometryArray.BY_REFERENCE );

		// the referenced data can be read and written
		//quadParts.setCapability(GeometryArray.ALLOW_REF_DATA_READ);
		//quadParts.setCapability(GeometryArray.ALLOW_REF_DATA_WRITE);

		createGeometry();
		setAppearance(BloodAppearance.APP());
	}


	private void createGeometry() {
	/* Create the float arrays for the coords, vels, accs, 
	   normals, and texture coords. Assign the coords, normals
	   and texture coords to the QuadArray using BY_REFERENCE.

	   Only the coords will be changed later.
	*/ 
		cs = new float[numPoints*3]; // to store each (x,y,z)
		vels = new float[numPoints*3];
		accs = new float[numPoints*3];
		norms = new float[numPoints*3];

		tcoords = new float[numPoints*2]; // two texture coords per coord
		// textureCoordsArray = new TexCoord2f[numPoints];

		// initialise position, vel, and acc
		// steps of 12 == 4 (x,y,z) coords == 1 quad at a time
		for(int i=0; i < numPoints*3; i=i+12)
			initQuadParticle(i);
		// refer to the coordinates in the QuadArray
		quadParts.setCoordRefFloat(cs); // use BY_REFERENCE

		/* Initialise texture coords. 
			 We are applying the texture to each quad in the array
			 For each quad (4 points):
			 { 0.0f, 0.0f, 1.0f, 0.0f,
			   1.0f, 1.0f, 0.0f, 1.0f };
			 One texture coords (s,t) for each point (x,y,z).
		*/
		for(int i=0; i < numPoints*2; i=i+8) {
			tcoords[i]   = 0.0f; tcoords[i+1] = 0.0f; // for 1 point
			tcoords[i+2] = 1.0f; tcoords[i+3] = 0.0f;
			tcoords[i+4] = 1.0f; tcoords[i+5] = 1.0f;
			tcoords[i+6] = 0.0f; tcoords[i+7] = 1.0f;
		}
		quadParts.setTexCoordRefFloat(0, tcoords); // use BY_REFERENCE

		// initialise normals to face in a general positive z-axis direction
		Vector3f norm = new Vector3f();
		for(int i=0; i < numPoints*3; i=i+3) {
			randomNormal(norm);
			norms[i] = norm.x; norms[i+1] = norm.y; norms[i+2] = norm.z;
		}
		quadParts.setNormalRefFloat(norms); // use BY_REFERENCE

		setGeometry(quadParts);
	}

	private void initQuadParticle(int i) {
	// initialise quad made up of 4 particles from i to i+11 
		setCoord(cs, i, p1); // set starting points
		setCoord(cs, i+3, p2);
		setCoord(cs, i+6, p3);
		setCoord(cs, i+9, p4);

		// random velocity in XZ plane with combined vector XZ_VELOCITY
		double xvel = Math.random()*XZ_VELOCITY;
		double zvel = Math.sqrt((XZ_VELOCITY*XZ_VELOCITY) - (xvel*xvel));
		vels[i] = (float)((Math.random() < 0.5) ? -xvel : xvel);   // x vel
		vels[i+2] = (float)((Math.random() < 0.5) ? -zvel : zvel); // z vel
		// y velocity
		vels[i+1] = (float)(Math.random() * Y_VELOCITY);
		
		// unchanging accelerations, downwards in y direction
		accs[i] = 0.0f; accs[i+1] = -GRAVITY; accs[i+2] = 0.0f;
 
		// all the particles in the quad have the same vel and acc
		copyCoord(vels, i+3, i);
		copyCoord(vels, i+6, i);
		copyCoord(vels, i+9, i);

		copyCoord(accs, i+3, i);
		copyCoord(accs, i+6, i);
		copyCoord(accs, i+9, i);
	}


	private void setCoord(float[] fs, int i, float[] p) {
	// copy the 3 floats in p[] into fs starting at i
		fs[i] = p[0];
		fs[i+1] = p[1];
		fs[i+2] = p[2];
	}


	private void copyCoord(float[] fs, int to, int from) {
	// copy 3 floats starting at fs[from] to fs[to]
		fs[to] = fs[from];
		fs[to+1] = fs[from+1];
		fs[to+2] = fs[from+2];
	}


	private void randomNormal(Vector3f v) {
	// Create a unit vector. The x- and y- values can be +ve or -ve.
	// The x-value is positive, so facing towards the viewer.
		float z = (float) Math.random();            // between 0-1
		float x = (float)(Math.random()*2.0 - 1.0); // -1 to 1
		float y = (float)(Math.random()*2.0 - 1.0); // -1 to 1
		v.set(x,y,z);
		v.normalize();
	}


		public boolean updateData( float offset ) {
		/* An update of the geometry is triggered by the system.
			 Rather than use geo, we directly access the cs[] array
			 back in QuadParticles. */
			// step in 12's == 4 (x,y,z) coor
			boolean bleeding= false;
			for(int i=0; i < numPoints*3; i=i+12) {
				if( ((cs[i+1] > offset) && (cs[i+4] > offset) && (cs[i+7] > offset) && (cs[i+10] > offset)) || (vels[1] > 0) ) {
					bleeding= true;
					for( int n=0; n<12; n++ ) {
						// Calculate new position and velocity (treating is as a projectile). The acceleration is constant.
						cs[i+n] += vels[i+n] * TIMESTEP + 0.5 * accs[i+n] * TIMESTEP * TIMESTEP;
						// alter the (x,y,z) values a bit, without this each quad stays a fixed size since each (x,y,z) coord is moved by the same amount
						cs[i+n]+= ((float)(Math.random()*DELTA*2.0f))-DELTA;
						// calculate new velocities
						vels[i+n] += accs[i+n] * TIMESTEP;
					}
				}
			}
			if( !bleeding ) for( int i=0; i < numPoints*3; i=i+12 ) initQuadParticle(i);
			return bleeding;
		}

	public static class BloodAppearance {
		private final static String TEX_FNM = "data/images/blood.gif";
		private final static Color3f black = new Color3f(0.0f, 0.0f, 0.0f);
		private final static Color3f white = new Color3f(.4f, .4f, .4f);
		private final static Color3f red = new Color3f(0.6f, 0.0f, 0.1f);
		private final static Color3f darkRed = new Color3f(.3f, 0.0f, 0.0f);
		private static Appearance app;

		public BloodAppearance() {}

		public static Appearance APP() {
			if(app!=null) return app;
			app = new Appearance();

			// blended transparency so texture can be irregular
			TransparencyAttributes tra = new TransparencyAttributes();
			tra.setTransparencyMode( TransparencyAttributes.BLENDED );
			app.setTransparencyAttributes( tra );

			// mix the texture and the material colour
			TextureAttributes ta = new TextureAttributes();
			ta.setTextureMode(TextureAttributes.MODULATE);
			app.setTextureAttributes(ta);

			// load and set the texture
			URL url = BloodAppearance.class.getResource(TEX_FNM);
			TextureLoader loader = new TextureLoader(url, null);
			Texture2D texture = (Texture2D) loader.getTexture();
			app.setTexture(texture); // set the texture

			// set the material: bloody
			Material mat = new Material(darkRed, black, red, white, 20.f);
			mat.setLightingEnable(true);
			app.setMaterial(mat);
			return app;
		}
	}
}
